#include <fstream>
#include <iostream>
#include <string>
#include <cstring>
#include <vector>
#include <limits>
#include <iomanip>

int romfix(const std::string& rawinfn);
bool getBinFile(std::ifstream& in, std::vector<std::uint8_t>& out);
size_t binFileSize(std::ifstream& in);

int main(int argc, char** argv) {
  if(argc >= 2) {
    if(strcmp(argv[1], "-h") == 0) {
      std::cout << "directions: list filenames without extension as arguments\n";
      return 0;
    }
  }
  for(int i=1;i<argc;i++) {
    std::string inn = argv[i];
    return romfix(inn);
  }

}

bool getBinFile(std::ifstream& in, std::vector<std::uint8_t>& out) {
	if(!in.is_open())
		return false;
	auto size = binFileSize(in);
	out.resize(size);
	in.read(reinterpret_cast<char*>(out.data()), size);
	return true;
}

size_t binFileSize(std::ifstream& in) {
	if(!in.is_open())
		return false;
	in.ignore( std::numeric_limits<std::streamsize>::max() );
	std::streamsize length = in.gcount();
	in.clear();   //  Since ignore will have set eof.
	in.seekg( 0, std::ios_base::beg );
	return length;
}

int romfix(const std::string& rawinfn) {
  std::string infn = rawinfn;
  std::string outfn = rawinfn + ".nes";

  std::clog << "Converting file '" << infn << "'\n";


  std::ifstream in{infn, std::ios::binary};
  std::ofstream out;

  std::vector<std::uint8_t> data; //yes, a vector is inefficient here, but I needed to guarantee continuous data

  getBinFile(in, data);

  //excise first 0x30 and last 0x20 bytes from data

  std::vector<std::uint8_t>::const_iterator first = data.begin() + 0x30;
  std::vector<std::uint8_t>::const_iterator last = data.end() - 0x20;
  std::vector<std::uint8_t> result(first, last);

  //adjust first line (if NES)
  //check if famicom or nes:
  if(strncmp(reinterpret_cast<const char*>(result.data()), "\x01*NINTENDO-HVC*", 15) == 0) {
    //famicom
    std::clog << "Famicom Game Detected\n";


    //clean crc spots from blocks 1&2

    result.erase(result.begin()+0x38, result.begin()+0x3a); //block 1 crc
    result.erase(result.begin()+0x3a, result.begin()+0x3c); //block 2 crc

    if(result[0x3a] != 0x03) {
      std::clog << "Error: Block 3 not located; Found byte 0x"<<std::setw(2);
      std::clog << std::setfill('0') << +result[0x3a] << ", expected 0x03.\n";
      return 3;
    }

    int fileNum = result[0x39];
    
    int fileStart = 0x3a;
    for(int i=0;i<fileNum;i++) {
      std::clog << "Processing file " << i << std::hex << " " << std::setfill('0');
      std::clog << "(beginning at 0x"<<std::setw(2)<<fileStart<<")\n";
      //find CRC address for file header block, remove it

      result.erase(result.begin()+fileStart+0x10, result.begin()+fileStart+0x12); //block 3 crc

      int fileSize = result[fileStart + 0xD] | (result[fileStart+0xE] << 8);
      //I don't like little-endian

      std::clog << "File size is 0x"<<std::setw(4)<<fileSize<< " bytes\n";


      if(result[fileStart + 0x10] != 0x04) {
        std::clog << "Error: Block 4 not located (checked 0x"<<fileStart+0x10<<"), ";
        std::clog << "expected 0x04, found 0x"<<std::setfill('0')<<std::setw(2);
        std::clog << +result[fileStart+0x10] << ")\n";
        
        break;
      }
      //get file size to find end of file data block
      //remove CRC for file data block
      
      result.erase(result.begin()+fileStart+0x11+fileSize, result.begin()+fileStart+0x13+fileSize);

      // increment fileStart for next file
      fileStart = fileStart+0x11+fileSize;
    }


    //find end of data, excise footer
    //end is a multiple of 0xffdc
    int end = 0xffdc * int(data.size() / 0xffdc);
    std::vector<std::uint8_t>::iterator fSt = result.begin() + end; 
    result.erase(fSt, result.end());

    //validate length, pad if necessary
    if(result.size() % 65500) {
      std::cout << "Length doesn't seem quite right...\n";
      return 4;
    }

    //change filename to represent it being a fds game
    outfn = rawinfn + ".fds";
    out.open(outfn, std::ios::binary);
  }
  else if(strncmp(reinterpret_cast<const char*>(result.data()), "NES", 3)) {
    std::clog << "NES Game detected\n";
    //repair invalid header byte
    result[3] = 0x1A;
    out.open(outfn, std::ios::binary);
  }
  else {
    std::clog << "Invalid game detected\n";
    return 1;
  }


  std::clog << "Original file size: 0x" << data.size() << "; Output file size: 0x" << result.size() << "\n";
  //write new vector to outfn

  out.write(reinterpret_cast<const char*>(result.data()), result.size());

  return 0;
}
